#!/usr/bin/env python

#################################################################
#
#    SVqcVCF
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import json
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    General stats for variants
#################################################################
def get_stats(vnt_obj, stat_dict, ID_list):
    ''' extract information from variant for single samples,
    update counts for each sample in ID_list '''
    var_type = variant_type_sv(vnt_obj)
    if var_type in ['del', 'dup']:
        for ID in ID_list:
            _genotype(vnt_obj, ID, var_type, stat_dict)
        #end for
    #end if
#end def

def _genotype(vnt_obj, ID, var_type, stat_dict):
    ''' genotype information, update counts for ID '''
    GT = vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
    if GT not in ['0/0', './.']: # sample has variant
        stat_dict[ID][var_type]['total'] += 1
    #end if
#end def

def to_json(stat_dict):
    ''' '''
    stat_json = {
        'total variants': []
    }

    for ID in stat_dict:
        tmp_total = {
            'name': ID,
            'total': 0
        }
        for k, v in stat_dict[ID].items():
            tmp_dict = {}
            # total variants
            if k in ['del', 'dup']:
                tmp_total.setdefault(k.upper(), v['total'])
                tmp_total['total'] += v['total']
            #end if
        #end for
        stat_json['total variants'].append(tmp_total)
    #end for
    return stat_json
#end def

#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    is_verbose = True if args['verbose'] else False
    stat_dict = {}

    # Buffers
    fo = open(args['outputfile'], 'w')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Get list of sample IDs to use
    ID_list = args['samples'] # list of sample IDs

    # Initializing stat_dict
    for ID in ID_list:
        stat_dict.setdefault(ID, {
                            'del': {'total': 0},
                            'dup': {'total': 0}
                            })
    #end for

    # Reading variants
    analyzed = 0
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if
        analyzed += 1
        # Getting and updating stats
        get_stats(vnt_obj, stat_dict, ID_list)
    #end for

    # Writing output
    sys.stderr.write('\n\n...Writing results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Create json
    stat_json = to_json(stat_dict)

    # Write json to file
    json.dump(stat_json, fo, indent=2, sort_keys=False)

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    shared_functions
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import bitarray
import tabix
import h5py
import numpy


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    Functions to buffer
#################################################################
def tabix_IT(filename, region):
    ''' open buffer to bgzip indexed filename using tabix,
    return an iterator to file content (tsv rows as lists)
    for region '''
    tb = tabix.open(filename)
    return tb.querys(region)
#end def

#################################################################
#    Functions to load
#################################################################
def load_big(filename):
    ''' read big filename into bitarrays_dict with the following
    structure {key: bitarray, ...} '''
    big = h5py.File(filename, 'r')
    bitarrays_dict = {k: bitarray.bitarray() for k in big.keys()}
    for k in big.keys():
        bitarrays_dict[k].frombytes(big[k][:].tostring())
    #end for
    big.close()
    return bitarrays_dict
#end def

def bed_to_bitarray(filename):
    ''' read bed filename into bitarrays_dict with the following
    structure {chrID: bitarray, ...} '''
    chr_tmp, pos_tmp = '', set()
    bitarrays_dict = {}
    with open(filename) as fi:
        for line in fi:
            line = line.rstrip().split('\t')
            if len(line) >= 3: # valid line for position in bed format
                chr, start, end = line[0], int(line[1]), int(line[2])
                if not chr_tmp: chr_tmp = chr
                #end if
                # Set positions in bitarray for current chromosome and reset data structures for next chromosome
                if chr_tmp != chr: # next chromosome
                    bitarrays_dict.setdefault(chr_tmp, bitarray.bitarray(max(pos_tmp) + 1))
                    bitarrays_dict[chr_tmp].setall(False)
                    for i in pos_tmp:
                        bitarrays_dict[chr_tmp][i] = True
                    #end for
                    chr_tmp, pos_tmp = chr, set()
                #end if
                # Adding new positions
                i = 0
                while (start + i) < end:
                    pos_tmp.add(start + i + 1) # +1 to index by one
                                               # 'The first base in a chromosome is numbered 0'
                                               # 'The end position in each BED feature is one-based'
                    i += 1
                #end while
            #end if
        #end for
        # Set positions into bitarray for last chromosome
        bitarrays_dict.setdefault(chr_tmp, bitarray.bitarray(max(pos_tmp) + 1))
        bitarrays_dict[chr_tmp].setall(False)
        for i in pos_tmp:
            bitarrays_dict[chr_tmp][i] = True
        #end for
    #end with
    return bitarrays_dict
#end def

#################################################################
#    Functions to write
#################################################################
def bitarray_tofile(bit_array, filename):
    ''' convert bit_array (bitarray) to bytes and write to filename '''
    with open(filename, 'wb') as fo:
        bit_array.tofile(fo)
    #end with
#end def

#################################################################
#    Functions to check
#################################################################
def check_region(region, chr_dict):
    ''' check if chromosme and region format are valid,
    chr_dict follow the structure {chrID: ..., ...} '''
    # Parse and check if region is valid
    if ':' in region:
        try:
            chr, strt_end = region.split(':')
            strt, end = map(int, strt_end.split('-'))
            if strt >= end:
                raise ValueError('\nERROR in parsing region, in region {0} starting index is larger than ending index\n'
                        .format(region))
            #end if
        except Exception:
            raise ValueError('\nERROR in parsing region, region {0} format is not recognized\n'
                    .format(region))
        #end try
    else:
        chr = region
    #end if
    # Check if chr is valid
    if not chr in chr_dict:
        raise ValueError('\nERROR in parsing region, {0} is not a valid chromosome format\n'
                .format(chr))
    #end if
#end def

def check_chrom(chrom):
    ''' check if chromosome is canonical and in a valid format '''
    chrom_repl = chrom.replace('chr', '')
    if chrom_repl in {'M', 'MT', 'X', 'Y'}:
        return True
    else:
        try:
            int_chrom_repl = int(chrom_repl)
        except Exception:
            return False
        #end try
        if int_chrom_repl > 0 and int_chrom_repl < 23:
            return True
        #end if
    #end if
    return False
#end def

def check_VEP(vnt_obj, idx, VEPremove, VEPrescue, VEPtag, sep='&'):
    ''' check VEP annotations from VEPtag '''
    try: val_get = vnt_obj.get_tag_value(VEPtag)
    except Exception: return False
    #end try
    trscrpt_list = val_get.split(',')
    # Get all terms
    for trscrpt in trscrpt_list:
        trscrpt_terms = set(trscrpt.split('|')[idx].split(sep))
        if trscrpt_terms.intersection(VEPrescue):
            return True
        elif trscrpt_terms.difference(VEPremove):
            return True
        #end if
    #end for
    return False
#end def

def check_spliceAI(vnt_obj, idx_list, SpAItag_list, thr=0.8):
    ''' check if SpliceAI max delta score value is over threshold thr,
    either access the max delta score directly from a tag,
    or calculate as the max of the different delta scores '''
    # if SpliceAI is within VEP annotations,
    # this fetch only the first transcript
    # and assumes all the transcripts share the same scores
    # Get max delta score
    if len(SpAItag_list) == 1: # max score from tag
        try:
            val_get = vnt_obj.get_tag_value(SpAItag_list[0])
            max_DS = float(val_get.split('|')[idx_list[0]])
        except Exception: return False
        #end try
    else: # default, multiple delta scores
        DS_list = []
        for i, SpAItag in enumerate(SpAItag_list):
            try:
                # if SpliceAI is within VEP and is at the end of Format
                # need to remove , that separate next transcript
                val_get = vnt_obj.get_tag_value(SpAItag)
                DS = val_get.split('|')[idx_list[i]]
                DS_list.append(float(DS.split(',')[0]))
            except Exception: return False
            #end try
        #end for
        max_DS = max(DS_list)
    #end if
    # Check max delta score
    if max_DS >= thr:
        return True
    #end if
    return False
#end def

def check_CLINVAR(vnt_obj, CLN_idx, CLNtag, CLNSIG_idx=0, CLNSIGtag='', CLINVARonly=[]):
    ''' check if CLNtag is present, if CLINVARonly
    check if variant has specified tags or keywords in CLNSIGtag '''
    # if clinvar is within VEP annotations,
    # this fetch only the first transcript
    # and assumes all the transcripts share the same clinvar
    try: val_get = vnt_obj.get_tag_value(CLNtag)
    except Exception: return False
    #end try
    # Check CLN
    CLN = val_get.split('|')[CLN_idx]
    if not CLN: return False
    #end if
    # Check CLNSIG
    if CLINVARonly:
        val_get = vnt_obj.get_tag_value(CLNSIGtag)
        CLNSIG = val_get.split('|')[CLNSIG_idx]
        for term in CLINVARonly:
            if term.lower() in CLNSIG.lower():
                return True
            #end if
        #end for
        return False
    #end if
    return True
#end def

#################################################################
#    Functions to get info
#################################################################
def variant_type(REF, ALT):
    ''' return variant type as snv, ins, del '''
    if len(ALT.split(',')) > 1:
        return 'snv' # TO DECIDE WHAT TO DO, as snv for now
    elif len(REF) > len(ALT) or ALT in ['*', '-']:
        return 'del'
    elif len(REF) < len(ALT) or REF in ['*', '-']:
        return 'ins'
    #end if
    return 'snv'
#end def

def variant_type_ext(REF, ALT):
    ''' return variant type as snv, ins, del,
    mnv (multi-nucleotide), mav (multi-allelic) '''
    # multi-allelic variant
    if len(ALT.split(',')) > 1:
        return 'mav'
    #end if
    # normal variant
    if len(REF) < len(ALT) or REF in ['*', '-']:
        return 'ins'
    elif len(REF) > len(ALT) or ALT in ['*', '-']:
        return 'del'
    elif len(REF) == len(ALT) and len(REF) != 1:
        return 'mnv'
    #end if
    return 'snv'
#end def

def variant_type_sv(vnt_obj):
    ''' return variant type as ins, del,
    dup, inv, bnd (translocation), cvn (copy number variation) '''
    try:
        SVTYPE = vnt_obj.get_tag_value("SVTYPE")
    except Exception:
        raise ValueError('\nERROR in parsing vcf, variant at position {0} does not contain SVTYPE in INFO\n'
                .format(vnt_obj.CHROM + ":" + str(vnt_obj.POS)))
    #end try
    if SVTYPE == 'DEL':
        return 'del'
    elif SVTYPE == 'DUP':
        return 'dup'
    elif SVTYPE == 'INS':
        return 'ins'
    elif SVTYPE == 'INV':
        return 'inv'
    elif SVTYPE == 'BND':
        return 'bnd'
    elif SVTYPE == 'CNV':
        return 'cnv'
    else:
        raise ValueError('\nERROR in parsing vcf, variant at position {0} contains unexpected SVTYPE "{1}" in INFO\n'
                .format(vnt_obj.CHROM + ":" + str(vnt_obj.POS), SVTYPE))
    #end if
#end def

def allele_frequency(vnt_obj, aftag, idx=0):
    ''' return allele frequency for variant from aftag in INFO,
    return 0. if tag is missing or value is not a float '''
    try:
        return float(vnt_obj.get_tag_value(aftag).split('|')[idx])
    except Exception:
        return 0.
    #end try
#end def

def VEP_field(vnt_obj, idx, VEPtag):
    ''' return list of annotations at idx across all transcripts from VEPtag '''
    try: val_get = vnt_obj.get_tag_value(VEPtag)
    except Exception: return []
    #end try
    trscrpt_list = val_get.split(',')
    return [trscrpt.split('|')[idx] for trscrpt in trscrpt_list]
#end def

def get_tag_idx(vnt_obj, tag, idx=0):
    ''' return value from tag at idx in INFO for variant,
    return None if tag is missing '''
    try:
        return vnt_obj.get_tag_value(tag).split('|')[idx]
    except Exception:
        return None
    #end try
#end def

#################################################################
#    Functions to modify
#################################################################
def clean_VEP(vnt_obj, idx, VEPremove, VEPrescue, VEPtag, sep='&'):
    ''' clean VEP annotations from VEPtag '''
    try: val_get = vnt_obj.get_tag_value(VEPtag)
    except Exception: return False
    #end try
    trscrpt_clean = []
    trscrpt_list = val_get.split(',')
    # Check Consequence terms and clean transcripts
    for trscrpt in trscrpt_list:
        cnsquence_terms = []
        trscrpt_values = trscrpt.split('|')
        for term in trscrpt_values[idx].split(sep):
            if term in VEPrescue:
                cnsquence_terms.append(term)
            elif term not in VEPremove:
                cnsquence_terms.append(term)
            #end if
        #end for
        if cnsquence_terms:
            trscrpt_values[idx] = sep.join(cnsquence_terms)
            trscrpt_clean.append('|'.join(trscrpt_values))
        #end if
    #end for
    return ','.join(trscrpt_clean)
#end def

def clean_VEP_byfield(vnt_obj, idx, values, VEPtag):
    ''' clean VEP annotations from VEPtag using a set of values,
    if VEPtag field at idx not in values remove transcript '''
    try: val_get = vnt_obj.get_tag_value(VEPtag)
    except Exception: return False
    #end try
    trscrpt_clean = []
    trscrpt_list = val_get.split(',')
    # Check field and clean transcripts
    for trscrpt in trscrpt_list:
        if trscrpt.split('|')[idx] in values:
            trscrpt_clean.append(trscrpt)
        #end if
    #end for
    return ','.join(trscrpt_clean)
#end def



#!/usr/bin/env python

#################################################################
#
#    qcVCF
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import json
import statistics
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser
# pedigree_parser
from granite.lib import pedigree_parser


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    General stats for variants
#################################################################
def get_stats(vnt_obj, stat_dict, ID_list):
    ''' extract information from variant for single samples,
    update counts for each sample in ID_list '''
    var_type = variant_type_ext(vnt_obj.REF, vnt_obj.ALT)
    for ID in ID_list:
        _genotype(vnt_obj, ID, var_type, stat_dict)
        _read_depth_DP(vnt_obj, ID, stat_dict)
        _read_depth_RSTR(vnt_obj, ID, stat_dict)
        if var_type == 'snv':
            _substitution(vnt_obj, ID, vnt_obj.REF, vnt_obj.ALT, stat_dict)
        #end if
    #end for
#end def

def _genotype(vnt_obj, ID, var_type, stat_dict):
    ''' genotype information, update counts for ID '''
    GT = vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
    if GT not in ['0/0', './.']: # sample has variant
        alt, alt_ = GT.split('/')
        if alt == alt_:
            stat_dict[ID][var_type]['hom'] += 1
        else:
            stat_dict[ID][var_type]['het'] += 1
        #end if
        stat_dict[ID][var_type]['total'] += 1
    #end if
#end def

def _substitution(vnt_obj, ID, REF, ALT, stat_dict):
    ''' substitution information, update counts for ID '''
    GT = vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
    if GT not in ['0/0', './.']: # sample has variant
        stat_dict[ID]['sub'][REF + '_' + ALT] += 1
    #end if
#end def

def _read_depth_DP(vnt_obj, ID, stat_dict):
    ''' read depth information based on DP, update data for ID '''
    GT = vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
    try: DP = vnt_obj.get_genotype_value(ID, 'DP')
    except Exception: return # missing DP information for that ID, skip
    #end try
    if GT not in ['0/0', './.']:
        stat_dict[ID]['depth']['DP'].append(int(DP))
    #end if
#end def

def _read_depth_RSTR(vnt_obj, ID, stat_dict):
    ''' read depth information based on RSTR, update data for ID '''
    GT = vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
    try: RSTR = vnt_obj.get_genotype_value(ID, 'RSTR')
    except Exception: return # missing RSTR information for that ID, skip
    #end try
    if GT not in ['0/0', './.']:
        stat_dict[ID]['depth']['RSTR'].append(sum(map(int, RSTR.split(','))))
    #end if
#end def

#################################################################
#    Pedigree stats
#################################################################
def get_stats_pedigree(pedigree_obj, vnt_obj, stat_dict, ID_list):
    ''' extract information from variant for pedigree,
    update counts for each sample in ID_list '''
    var_type = variant_type_ext(vnt_obj.REF, vnt_obj.ALT)
    for ID in ID_list:
        _mendelian_error(pedigree_obj, vnt_obj, ID, var_type, stat_dict)
    #end for
#end def

def _mendelian_error(pedigree_obj, vnt_obj, ID, var_type, stat_dict):
    ''' check for mendelian error based on trio information,
    update counts for ID '''
    GT, parents_GT = vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/'), []
    if GT not in ['0/0', './.']: # sample has variant
        parents = pedigree_obj.get_member_by_sample(ID).get_parents()
        if len(parents) < 2: return # missing trio information, skip variant
        #end if
        for parent in parents:
            if not parent.is_sample(): # missing parent sample information,
                return                 # skip variant
            #end if
            try:
                GT_ = vnt_obj.get_genotype_value(parent.sample, 'GT').replace('|', '/')
            except Exception: return # parent sample missing in vcf, skip variant
            #end try
            if GT_ == './.':
                if GT in ['0/1', '1/0']:
                    stat_dict[ID]['trio'][var_type]['het']['missing_in_parent'] += 1
                    stat_dict[ID]['trio'][var_type]['het']['total'] += 1
                elif GT == '1/1':
                    stat_dict[ID]['trio'][var_type]['hom']['missing_in_parent'] += 1
                    stat_dict[ID]['trio'][var_type]['hom']['total'] += 1
                #end if
                return # missing genotype for parent, skip
            #end if
            parents_GT.append(GT_)
        #end for
        GT_0, GT_1 = parents_GT[0], parents_GT[1]
        # Check errors
        if GT in ['0/1', '1/0']:
            if GT_0 == '0/0' and GT_1 == '0/0':
                stat_dict[ID]['trio'][var_type]['het']['de_novo'] += 1
            elif GT_0 == '1/1' and GT_1 == '1/1':
                stat_dict[ID]['trio'][var_type]['het']['errors'] += 1
            #end if
            stat_dict[ID]['trio'][var_type]['het']['total'] += 1
        elif GT == '1/1':
            if GT_0 == '0/0' or GT_1 == '0/0':
                stat_dict[ID]['trio'][var_type]['hom']['errors'] += 1
            #end if
            stat_dict[ID]['trio'][var_type]['hom']['total'] += 1
        #end if
    #end if
#end def

#################################################################
#
#################################################################
def tt_ratio(sub_dict):
    ''' return transition-transversion ratio'''
    ti = sub_dict['A_G'] + sub_dict['G_A'] + sub_dict['T_C'] + sub_dict['C_T']
    tv = sub_dict['A_T'] + sub_dict['T_A'] + sub_dict['A_C'] + sub_dict['C_A'] \
        + sub_dict['G_T'] + sub_dict['T_G'] + sub_dict['G_C'] + sub_dict['C_G']
    return ti / tv
#end def

def compress_list(data_list):
    ''' return compressed version of list of numbers as list of single points
    accompanied by list with occurency counts for each of the points,
    [points_list, counts_list] '''
    points, counts = [], []
    for i, e in enumerate(sorted(data_list)):
        if i == 0:
            tmp_e, count = e, 1
        elif e != tmp_e:
            points.append(tmp_e)
            counts.append(count)
            tmp_e, count = e, 1
        else: count += 1
        #end if
    #end for
    # last point
    points.append(tmp_e)
    counts.append(count)
    return [points, counts]
#end def

def to_json(stat_dict, stat_to_add):
    ''' '''
    stat_json = {
        'total variants': [],
        'depth of coverage': []
    }
    if 'ti_tv' in stat_to_add:
        stat_json.setdefault('transition-transversion ratio', [])
    #end if
    if 'het_hom' in stat_to_add:
        stat_json.setdefault('heterozygosity ratio', {
                                               'SNV': [], 'INS': [],
                                               'DEL': [], 'MNV': []
                                           })
    #end if
    if 'trio_errors' in stat_to_add:
        stat_json.setdefault('mendelian errors in trio', {
                                               'SNV': [], 'INS': [],
                                               'DEL': []
                                           })
    #end if
    for ID in stat_dict:
        tmp_total = {
            'name': ID,
            'total': 0
        }
        for k, v in stat_dict[ID].items():
            tmp_dict = {}
            # total variants
            if k in ['snv', 'ins', 'del', 'mnv', 'mav']:
                tmp_total.setdefault(k.upper(), v['total'])
                tmp_total['total'] += v['total']
            #end if
            # heterozygosity ratio
            if k in ['snv', 'ins', 'del', 'mnv'] and 'het_hom' in stat_to_add:
                tmp_dict.setdefault('name', ID)
                if v['hom']:
                    hh_ratio = round(v['het'] / v['hom'], 2)
                    tmp_dict.setdefault('ratio', hh_ratio)
                #end if
                tmp_dict.setdefault('counts', v)
                stat_json['heterozygosity ratio'][k.upper()].append(tmp_dict)
            #end if
            # substitutions
            if k == 'sub' and 'ti_tv' in stat_to_add:
                tmp_dict.setdefault('name', ID)
                try: tmp_dict.setdefault('ratio', round(tt_ratio(v), 2))
                except Exception: pass
                #end try
                tmp_dict.setdefault('counts', v)
                stat_json['transition-transversion ratio'].append(tmp_dict)
            #end if
            # mendelian errors trio
            if k == 'trio' and 'trio_errors' in stat_to_add:
                for k_v, v_v in v.items():
                    tmp_dict = {} # reset tmp_dict
                    if k_v in ['snv', 'ins', 'del']:
                        if v_v['het']['total'] or v_v['hom']['total']:
                            tmp_dict.setdefault('name', ID)
                            tmp_dict.setdefault('counts', v_v)
                            stat_json['mendelian errors in trio'][k_v.upper()].append(tmp_dict)
                        #end if
                    #end if
                #end for
            #end if
            # depth of coverage
            if k == 'depth':
                tmp_dict.setdefault('name', ID)
                if v['DP']: # DP
                    tmp_dict.setdefault('DP (gatk)', {})
                    tmp_dict['DP (gatk)'].setdefault('average', round(statistics.mean(v['DP']), 2))
                    tmp_dict['DP (gatk)'].setdefault('points', compress_list(v['DP']))
                #end if
                if v['RSTR']: # RSTR
                    tmp_dict.setdefault('DP (raw)', {})
                    tmp_dict['DP (raw)'].setdefault('average', round(statistics.mean(v['RSTR']), 2))
                    tmp_dict['DP (raw)'].setdefault('points', compress_list(v['RSTR']))
                #end if
                stat_json['depth of coverage'].append(tmp_dict)
            #end if
        #end for
        stat_json['total variants'].append(tmp_total)
    #end for
    return stat_json
#end def

#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    is_verbose = True if args['verbose'] else False
    stat_dict = {}
    stat_to_add = []

    # Check command line arguments
    if args['ti_tv']: stat_to_add.append('ti_tv')
    #end if
    if args['het_hom']: stat_to_add.append('het_hom')
    #end if
    if args['trio_errors']: stat_to_add.append('trio_errors')
    #end if
    if not stat_to_add:
        sys.exit('\nERROR in parsing arguments: specify at least one metric to add\n')
    #end if

    # Buffers
    fo = open(args['outputfile'], 'w')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Get list of sample IDs to use
    ID_list = args['samples'] # list of sample IDs

    # Creating Pedigree object
    pedigree_obj = pedigree_parser.Pedigree(args['pedigree'])

    # Initializing stat_dict
    for ID in ID_list:
        stat_dict.setdefault(ID, {
                            'snv': {'het': 0, 'hom': 0, 'total': 0},
                            'ins': {'het': 0, 'hom': 0, 'total': 0},
                            'del': {'het': 0, 'hom': 0, 'total': 0},
                            'mnv': {'het': 0, 'hom': 0, 'total': 0},
                            'mav': {'het': 0, 'hom': 0, 'total': 0},
                            'sub': {
                                'A_G': 0, 'A_T': 0, 'A_C': 0,
                                'T_A': 0, 'T_G': 0, 'T_C': 0,
                                'C_A': 0, 'C_G': 0, 'C_T': 0,
                                'G_A': 0, 'G_T': 0, 'G_C': 0
                                },
                            'trio': {
                                'snv': {
                                        'het': {'de_novo': 0, 'errors': 0, 'missing_in_parent': 0, 'total': 0},
                                        'hom': {'errors': 0, 'missing_in_parent': 0, 'total': 0}
                                        },
                                'ins': {
                                        'het': {'de_novo': 0, 'errors': 0, 'missing_in_parent': 0, 'total': 0},
                                        'hom': {'errors': 0, 'missing_in_parent': 0, 'total': 0}
                                        },
                                'del': {
                                        'het': {'de_novo': 0, 'errors': 0, 'missing_in_parent': 0, 'total': 0},
                                        'hom': {'errors': 0, 'missing_in_parent': 0, 'total': 0}
                                        },
                                'mnv': {
                                        'het': {'de_novo': 0, 'errors': 0, 'missing_in_parent': 0, 'total': 0},
                                        'hom': {'errors': 0, 'missing_in_parent': 0, 'total': 0}
                                        },
                                'mav': {
                                        'het': {'de_novo': 0, 'errors': 0, 'missing_in_parent': 0, 'total': 0},
                                        'hom': {'errors': 0, 'missing_in_parent': 0, 'total': 0}
                                        }
                                },
                            'depth': {
                                'RSTR': [],
                                'DP': []
                            }
                            })
    #end for

    # Reading variants
    analyzed = 0
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM):
        #     continue
        # #end if
        analyzed += 1

        # Getting and updating stats
        get_stats(vnt_obj, stat_dict, ID_list)
        get_stats_pedigree(pedigree_obj, vnt_obj, stat_dict, ID_list)
    #end for

    # Writing output
    sys.stderr.write('\n\n...Writing results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Create json
    stat_json = to_json(stat_dict, stat_to_add)

    # Write json to file
    json.dump(stat_json, fo, indent=2, sort_keys=False)

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    comHet
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import json
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser
# shared_vars
from granite.lib.shared_vars import VEP_encode
from granite.lib.shared_vars import DStags


#################################################################
#
#    OBJECTS
#
#################################################################
class VariantHet(object):
    ''' object to extend a Variant object to store additional
    information on genes, transcripts and compound heterozygous pairs '''

    def __init__(self, vnt_obj, i):
        ''' initialize VariantHet object '''
        self.comHet = []
        self.vnt_obj = vnt_obj
        self.ENST_dict = {} # {ENSG: ENST_set, ...}
        self.ENSG_IMPCT_dict = {} # {ENSG: IMPACT, ...}
        self.ENST_IMPCT_dict = {} # {ENST: IMPACT, ...}
        self.i = i # variant index
        self.SpAI = ''
        self.CLINVAR = ''
    #end def

    def add_SpAI(self, SpAI_val, SpAI_thr=0.8):
        ''' add SpliceAI, encoded as S or s based on SpAI_thr '''
        if SpAI_val >= SpAI_thr: self.SpAI = 'S'
        else: self.SpAI = 's'
        #end if
    #end def

    def add_CLINVAR(self, CLNSIG_val, CLNSIG_encode):
        ''' add CLINVAR, encoded as C or c based on CLNSIG value '''
        for clnsig, encd in CLNSIG_encode:
            if clnsig in CLNSIG_val:
                self.CLINVAR = encd
                break
            #end if
        #end for
    #end def

    def add_ENST(self, ENSG, ENST_set):
        ''' add gene and its associated transcripts information '''
        self.ENST_dict.setdefault(ENSG, ENST_set)
    #end def

    def add_ENST_IMPCT(self, ENST, IMPCT):
        ''' add transcript and its associated IMPACT,
        IMPACT encoded as number '''
        self.ENST_IMPCT_dict.setdefault(ENST, IMPCT)
    #end def

    def add_ENSG_IMPCT(self, ENSG, IMPCT_set):
        ''' add gene and worst IMPACT in its associated transcripts,
        IMPACT encoded as number '''
        self.ENSG_IMPCT_dict.setdefault(ENSG, sorted(list(IMPCT_set))[0])
    #end def

    def get_gene_impct(self, ENSG):
        ''' return combined impact information for gene '''
        return self.ENSG_IMPCT_dict[ENSG], self.SpAI + self.CLINVAR
    #end def

    def get_trscrpt_impct(self, ENST):
        ''' return combined impact information for transcript '''
        return self.ENST_IMPCT_dict[ENST], self.SpAI + self.CLINVAR
    #end def

    def map_impct(self, impct, impct_, IMPCT_decode, test):
        ''' '''
        if test:
            return IMPCT_decode[impct[0]] + impct[1] + '/' + IMPCT_decode[impct_[0]] + impct_[1]
        else:
            STRONG_set = {'H', 'M', 'S', 'C'}
            impct_set = set(list(IMPCT_decode[impct[0]] + impct[1]))
            impct_set_ = set(list(IMPCT_decode[impct_[0]] + impct_[1]))
            if impct_set.intersection(STRONG_set) and \
               impct_set_.intersection(STRONG_set):
                return 'STRONG_PAIR'
            elif impct_set.intersection(STRONG_set) or \
                 impct_set_.intersection(STRONG_set):
                return 'MEDIUM_PAIR'
            else:
                return 'WEAK_PAIR'
            #end if
        #end if
    #end def

    def worse_ENST(self, vntHet_obj, common_ENST):
        ''' return the shared transcript with worse IMPACT '''
        score, tmp_score, tmp_ENST = 0, 0, ''
        for ENST in common_ENST:
            score = self.ENST_IMPCT_dict[ENST] + vntHet_obj.ENST_IMPCT_dict[ENST]
            if not tmp_score:
                tmp_ENST = ENST
                tmp_score = score
            else:
                if score < tmp_score:
                    tmp_ENST = ENST
                #end if
            #end if
        #end for
        return tmp_ENST
    #end def

    def add_pair(self, vntHet_obj, ENSG, phase, sep, is_impct, IMPCT_decode, test):
        ''' add information for compound heterozygous pair with vntHet_obj '''
        comHet_pair = [phase, ENSG]
        common_ENST = self.ENST_dict[ENSG].intersection(vntHet_obj.ENST_dict[ENSG])
        if common_ENST:
            comHet_pair.append(sep.join(sorted(common_ENST)))
        else: comHet_pair.append('')
        #end if
        if is_impct:
            # add gene impact
            impct, impct_ = sorted([self.get_gene_impct(ENSG), vntHet_obj.get_gene_impct(ENSG)])
            comHet_pair.append(self.map_impct(impct, impct_, IMPCT_decode, test))
            # add transcript impact
            if common_ENST:
                ENST = self.worse_ENST(vntHet_obj, sorted(common_ENST))
                impct, impct_ = sorted([self.get_trscrpt_impct(ENST), vntHet_obj.get_trscrpt_impct(ENST)])
                comHet_pair.append(self.map_impct(impct, impct_, IMPCT_decode, test))
            else: comHet_pair.append('')
            #end if
        #end if
        comHet_pair.append(vntHet_obj.vnt_obj.repr())
        self.comHet.append('|'.join(comHet_pair))
    #end def

    def to_string(self):
        ''' return variant as a string after adding comHet information to INFO field '''
        if self.comHet:
            self.vnt_obj.add_tag_info('comHet=' + ','.join(self.comHet))
        #end if
        return self.vnt_obj.to_string()
    #end def

#end class VariantHet


#################################################################
#
#    FUNCTIONS
#
#################################################################
def is_comHet(vntHet_obj_1, vntHet_obj_2, ID_list, allow_undef=False):
    ''' check genotypes combination for parents if available and
    refine the assignment for the pair as a compound heterozygous or not '''
    for ID in ID_list[1:]:
        GT_1 = vntHet_obj_1.vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
        GT_2 = vntHet_obj_2.vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
        if GT_1 == '1/1' or GT_2 == '1/1':
            return False
        elif GT_1 in ['0/1', '1/0'] and GT_2 in ['0/1', '1/0']:
            return False
        #end if
        if not allow_undef:
            if GT_1 == './.' or GT_2 == './.':
                return False
            #end if
        #end if
    #end for
    return True
#end def

def phase(vntHet_obj_1, vntHet_obj_2, ID_list):
    ''' check genotypes combination for parents if available and
    refine the assignment for the pair as Phased or Unphased '''
    if len(ID_list[1:]) < 2:
        return 'Unphased'
    #end if
    for ID in ID_list[1:]:
        GT_1 = vntHet_obj_1.vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
        GT_2 = vntHet_obj_2.vnt_obj.get_genotype_value(ID, 'GT').replace('|', '/')
        if GT_1 == '0/0' and GT_2 == '0/0': # this could be a potential de novo
                                            # in a compound het
                return 'Unphased'
        elif GT_1 == './.' or GT_2 == './.':
                return 'Unphased'
        #end if
    #end for
    return 'Phased'
#end def

def encode_IMPACT(IMPCT_list, VEP_encode, IMPCT_encode, is_IMPACT, sep):
    ''' return IMPCT_list encoded as number '''
    IMPCT_encoded = []
    if is_IMPACT: # encode IMPACT
        IMPCT_encoded = [IMPCT_encode[IMPCT] for IMPCT in IMPCT_list]
    else: # encode terms from Consequence
        for terms in IMPCT_list:
            impct_set = set()
            for term in terms.split(sep):
                try: impct_set.add(VEP_encode[term])
                except Exception: impct_set.add(VEP_encode['MODIFIER'])
                #end try
            #end for
            IMPCT_encoded.append(sorted(list(impct_set))[0])
        #end for
    #end if
    return IMPCT_encoded
#end def

def update_stats(vntHet_obj, stat_dict, sep, is_impct):
    ''' '''
    try: val_get = vntHet_obj.vnt_obj.get_tag_value('comHet')
    except Exception: return
    #end try
    var_repr = vntHet_obj.vnt_obj.repr()
    var_phase = 'Unphased'
    for cmpHet in val_get.split(','):
        if is_impct:
            PHASE, ENSG_ID, ENST_ID, IMPCT_G, IMPCT_T, VARIANT = cmpHet.split('|')
        else: PHASE, ENSG_ID, ENST_ID, VARIANT = cmpHet.split('|')
        #end if
        if PHASE == 'Phased': var_phase = PHASE
        #end if
        # global
        if (var_repr, VARIANT) not in stat_dict['pairs']['pairs_set']:
            # increase count
            stat_dict['pairs'][PHASE] += 1
            # add pair to set
            stat_dict['pairs']['pairs_set'].add((var_repr, VARIANT))
            stat_dict['pairs']['pairs_set'].add((VARIANT, var_repr))
        #end if
        # by gene
        stat_dict['genes'].setdefault(ENSG_ID, {
                                            'vntHet_set': {},
                                            'pairs_set': set(),
                                            'transcript_set': {},
                                            'Phased': 0,
                                            'Unphased': 0
                                            })
        stat_dict['genes'][ENSG_ID]['vntHet_set'].setdefault(vntHet_obj, 0)
        if PHASE == 'Phased':
            stat_dict['genes'][ENSG_ID]['vntHet_set'][vntHet_obj] = 1
        #end if
        if (var_repr, VARIANT) not in stat_dict['genes'][ENSG_ID]['pairs_set']:
            # increase count
            stat_dict['genes'][ENSG_ID][PHASE] += 1
            # add pair to set
            stat_dict['genes'][ENSG_ID]['pairs_set'].add((var_repr, VARIANT))
            stat_dict['genes'][ENSG_ID]['pairs_set'].add((VARIANT, var_repr))
        #end if
        # by transcripts
        if ENST_ID:
            for trscrpt in ENST_ID.split(sep):
                stat_dict['trscrpts'].setdefault(trscrpt, {
                                                    'gene': ENSG_ID,
                                                    'vntHet_set': {},
                                                    'pairs_set': set(),
                                                    'Phased': 0,
                                                    'Unphased': 0
                                                    })
                stat_dict['genes'][ENSG_ID]['transcript_set'].setdefault(trscrpt, 0)
                stat_dict['trscrpts'][trscrpt]['vntHet_set'].setdefault(vntHet_obj, 0)
                if PHASE == 'Phased':
                    stat_dict['genes'][ENSG_ID]['transcript_set'][trscrpt] = 1
                    stat_dict['trscrpts'][trscrpt]['vntHet_set'][vntHet_obj] = 1
                #end if
                if (var_repr, VARIANT) not in stat_dict['trscrpts'][trscrpt]['pairs_set']:
                    # increase count
                    stat_dict['trscrpts'][trscrpt][PHASE] += 1
                    # add pair to set
                    stat_dict['trscrpts'][trscrpt]['pairs_set'].add((var_repr, VARIANT))
                    stat_dict['trscrpts'][trscrpt]['pairs_set'].add((VARIANT, var_repr))
                #end if
            #end for
        #end if
        # by impact
        if is_impct:
            stat_dict['impact'].setdefault(IMPCT_G, {
                                                'vntHet_set': {},
                                                'pairs_set': set(),
                                                'gene_set': {},
                                                'transcript_set': {},
                                                'Phased': 0,
                                                'Unphased': 0
                                                    })
            stat_dict['impact'][IMPCT_G]['vntHet_set'].setdefault(vntHet_obj, 0)
            stat_dict['impact'][IMPCT_G]['gene_set'].setdefault(ENSG_ID, 0)
            if ENST_ID:
                for trscrpt in ENST_ID.split(sep):
                    stat_dict['impact'][IMPCT_G]['transcript_set'].setdefault(trscrpt, 0)
                    if PHASE == 'Phased':
                        stat_dict['impact'][IMPCT_G]['transcript_set'][trscrpt] = 1
                    #end if
                #end for
            #end if
            if PHASE == 'Phased':
                stat_dict['impact'][IMPCT_G]['vntHet_set'][vntHet_obj] = 1
                stat_dict['impact'][IMPCT_G]['gene_set'][ENSG_ID] = 1
            #end if
            if (var_repr, VARIANT) not in stat_dict['impact'][IMPCT_G]['pairs_set']:
                # increase count
                stat_dict['impact'][IMPCT_G][PHASE] += 1
                # add pair to set
                stat_dict['impact'][IMPCT_G]['pairs_set'].add((var_repr, VARIANT))
                stat_dict['impact'][IMPCT_G]['pairs_set'].add((VARIANT, var_repr))
            #end if
        #end if
    #end for
    stat_dict['vnts'][var_phase] += 1
#end def

def to_json(stat_dict, is_impct):
    ''' '''
    stat_json = {}
    phased_genes, total_genes = 0, 0
    phased_trscrpts, total_trscrpts = 0, 0
    # phased genes
    for ENSG_ID in stat_dict['genes']:
        if stat_dict['genes'][ENSG_ID]['Phased']:
            phased_genes += 1
        #end if
        total_genes += 1
    #end for
    # phased transcripts
    for trscrpt in stat_dict['trscrpts']:
        if stat_dict['trscrpts'][trscrpt]['Phased']:
            phased_trscrpts += 1
        #end if
        total_trscrpts += 1
    #end for
    # global stats
    stat_json.setdefault('general', {})
    stat_json['general'].setdefault('genes', {'phased': phased_genes, 'total': total_genes})
    stat_json['general'].setdefault('transcripts', {'phased': phased_trscrpts, 'total': total_trscrpts})
    stat_json['general'].setdefault('variants', {'phased': stat_dict['vnts']['Phased'],
                                                 'total': stat_dict['vnts']['Phased'] + stat_dict['vnts']['Unphased']})
    stat_json['general'].setdefault('pairs', {'phased': stat_dict['pairs']['Phased'],
                                              'total': stat_dict['pairs']['Phased'] + stat_dict['pairs']['Unphased']})
    # by genes
    stat_json.setdefault('by_genes', [])
    for ENSG_ID in sorted(stat_dict['genes']):
        variants_phased, variants_total = 0, 0
        for v, c in stat_dict['genes'][ENSG_ID]['vntHet_set'].items():
            if c == 1: variants_phased += 1
            #end if
            variants_total += 1
        #end for
        transcripts_phased, transcripts_total = 0, 0
        for v, c in stat_dict['genes'][ENSG_ID]['transcript_set'].items():
            if c == 1: transcripts_phased += 1
            #end if
            transcripts_total += 1
        #end for
        tmp_dict = {}
        tmp_dict.setdefault('name', ENSG_ID)
        tmp_dict.setdefault('transcripts', {'phased': transcripts_phased,
                                            'total': transcripts_total
                                            })
        tmp_dict.setdefault('variants', {'phased': variants_phased,
                                         'total': variants_total
                                        })
        tmp_dict.setdefault('pairs', {'phased': stat_dict['genes'][ENSG_ID]['Phased'],
                                      'total': stat_dict['genes'][ENSG_ID]['Phased'] + stat_dict['genes'][ENSG_ID]['Unphased']
                                     })
        # append to list
        stat_json['by_genes'].append(tmp_dict)
    #end for
    # by transcripts
    stat_json.setdefault('by_transcripts', [])
    for trscrpt in sorted(stat_dict['trscrpts']):
        variants_phased, variants_total = 0, 0
        for v, c in stat_dict['trscrpts'][trscrpt]['vntHet_set'].items():
            if c == 1: variants_phased += 1
            #end if
            variants_total += 1
        #end for
        tmp_dict = {}
        tmp_dict.setdefault('name', trscrpt)
        tmp_dict.setdefault('gene', stat_dict['trscrpts'][trscrpt]['gene'])
        tmp_dict.setdefault('variants', {'phased': variants_phased,
                                         'total': variants_total
                                        })
        tmp_dict.setdefault('pairs', {'phased': stat_dict['trscrpts'][trscrpt]['Phased'],
                                      'total': stat_dict['trscrpts'][trscrpt]['Phased'] + stat_dict['trscrpts'][trscrpt]['Unphased']
                                     })
        # append to list
        stat_json['by_transcripts'].append(tmp_dict)
    #end for
    # by impact
    if is_impct:
        stat_json.setdefault('by_impact', [])
        for impact in sorted(stat_dict['impact']):
            genes_phased, genes_total = 0, 0
            for v, c in stat_dict['impact'][impact]['gene_set'].items():
                if c == 1: genes_phased += 1
                #end if
                genes_total += 1
            #end for
            variants_phased, variants_total = 0, 0
            for v, c in stat_dict['impact'][impact]['vntHet_set'].items():
                if c == 1: variants_phased += 1
                #end if
                variants_total += 1
            #end for
            transcripts_phased, transcripts_total = 0, 0
            for v, c in stat_dict['impact'][impact]['transcript_set'].items():
                if c == 1: transcripts_phased += 1
                #end if
                transcripts_total += 1
            #end for
            tmp_dict = {}
            tmp_dict.setdefault('name', impact)
            tmp_dict.setdefault('genes', {'phased': genes_phased,
                                          'total': genes_total
                                         })
            tmp_dict.setdefault('transcripts', {'phased': transcripts_phased,
                                                'total': transcripts_total
                                                })
            tmp_dict.setdefault('variants', {'phased': variants_phased,
                                             'total': variants_total
                                            })
            tmp_dict.setdefault('pairs', {'phased': stat_dict['impact'][impact]['Phased'],
                                          'total': stat_dict['impact'][impact]['Phased'] + stat_dict['impact'][impact]['Unphased']
                                         })
            # append to list
            stat_json['by_impact'].append(tmp_dict)
        #end for
    #end if
    return stat_json
#end def

def print_stats(stat_json, fo, is_impct):
    ''' '''
    # global stats
    fo.write('##general stats\n')
    fo.write('#category\tphased\ttotal\n')
    fo.write('genes\t{0}\t{1}\n'.format(stat_json['general']['genes']['phased'],
                                        stat_json['general']['genes']['total']))
    fo.write('transcripts\t{0}\t{1}\n'.format(stat_json['general']['transcripts']['phased'],
                                              stat_json['general']['transcripts']['total']))
    fo.write('variants\t{0}\t{1}\n'.format(stat_json['general']['variants']['phased'],
                                           stat_json['general']['variants']['total']))
    fo.write('pairs\t{0}\t{1}\n'.format(stat_json['general']['pairs']['phased'],
                                        stat_json['general']['pairs']['total']))
    # by genes
    fo.write('\n##stats by genes\n')
    fo.write('#ENSG_ID\ttranscripts_phased\ttranscripts_total\tvariants_phased\tvariants_total\tpairs_phased\tpairs_total\n')
    for gene_dict in stat_json['by_genes']:
        fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n'.format(
                                            gene_dict['name'],
                                            gene_dict['transcripts']['phased'],
                                            gene_dict['transcripts']['total'],
                                            gene_dict['variants']['phased'],
                                            gene_dict['variants']['total'],
                                            gene_dict['pairs']['phased'],
                                            gene_dict['pairs']['total']
                                            ))
    #end for
    # by transcripts
    fo.write('\n##stats by transcripts\n')
    fo.write('#ENST_ID\tENSG_ID\tvariants_phased\tvariants_total\tpairs_phased\tpairs_total\n')
    for trscrpt_dict in stat_json['by_transcripts']:
        fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n'.format(
                                            trscrpt_dict['name'],
                                            trscrpt_dict['gene'],
                                            trscrpt_dict['variants']['phased'],
                                            trscrpt_dict['variants']['total'],
                                            trscrpt_dict['pairs']['phased'],
                                            trscrpt_dict['pairs']['total']
                                            ))
    #end for
    # by impact
    if is_impct:
        fo.write('\n##stats by impact\n')
        fo.write('#impact\tgenes_phased\tgenes_total\ttranscripts_phased\ttranscripts_total\tvariants_phased\tvariants_total\tpairs_phased\tpairs_total\n')
        for impact_dict in stat_json['by_impact']:
            fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\n'.format(
                                                impact_dict['name'],
                                                impact_dict['genes']['phased'],
                                                impact_dict['genes']['total'],
                                                impact_dict['transcripts']['phased'],
                                                impact_dict['transcripts']['total'],
                                                impact_dict['variants']['phased'],
                                                impact_dict['variants']['total'],
                                                impact_dict['pairs']['phased'],
                                                impact_dict['pairs']['total']
                                                ))
        #end for
    #end if
#end def

#################################################################
#    runner
#################################################################
def main(args, test=False):
    ''' '''
    # Definitions
    CLNSIG_encode = [
                      # high impact
                      ('Pathogenic', 'C'), ('Likely_pathogenic', 'C'),
                      # moderate/low impact
                      ('Conflicting_interpretations', 'c'), ('Uncertain_significance', 'c'), ('risk_factor', 'c')
                    ]
    # VEP_encode = {...} -> import from shared_vars
    # DStags = {...} -> import from shared_vars
    IMPCT_encode = {'HIGH': 1, 'MODERATE': 2, 'LOW': 3, 'MODIFIER': 4}
    IMPCT_decode = {1: 'H', 2: 'M', 3: 'L', 4: 'm'}

    # Variables
    is_impct = True if args['impact'] else False
    is_IMPACT = False # True if IMPACT field is in VEP
    is_family = False
    CLNSIGtag, CLNSIG_idx, is_CLNSIG = '', 0, False
    SpAItag_list, SpAI_idx_list, is_SpAI = [], [], False
    ENSG_idx, ENST_idx, IMPCT_idx = 0, 0, 0
    SpliceAItag = args['SpliceAItag'] # default None
    VEPtag = args['VEPtag'] if args['VEPtag'] else 'CSQ'
    sep = args['sep'] if args['sep'] else '&'
    allow_undef = True if args['allow_undef'] else False
    filter_cmpHet = True if args['filter_cmpHet'] else False
    granite_def = '##GRANITE=<ID=comHet>'
    comHet_def = '##INFO=<ID=comHet,Number=.,Type=String,Description="Putative compound heterozygous pairs. Subembedded:\'cmpHet\':Format:\'phase|gene|transcript|mate_variant\'">'
    comHet_impct_def = '##INFO=<ID=comHet,Number=.,Type=String,Description="Putative compound heterozygous pairs. Subembedded:\'cmpHet\':Format:\'phase|gene|transcript|impact_gene|impact_transcript|mate_variant\'">'
    is_verbose = True if args['verbose'] else False

    # Buffers
    fo = open(args['outputfile'], 'w', encoding='utf-8')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Add definition to header
    if is_impct:
        vcf_obj.header.add_tag_definition(granite_def + '\n' + comHet_impct_def, 'INFO')
    else:
        vcf_obj.header.add_tag_definition(granite_def + '\n' + comHet_def, 'INFO')
    #end if

    # Writing header
    vcf_obj.write_header(fo)

    # Data structures
    stat_dict = {'genes': {},
                 'trscrpts': {},
                 'pairs': {
                    'pairs_set': set(),
                    'Phased': 0,
                    'Unphased': 0
                    },
                 'vnts': {
                    'Phased': 0,
                    'Unphased': 0
                    },
                 'impact' : {}
                }
    ENSG_dict = {} # {ENSG: [vntHet_obj1, vntHet_obj2], ...}
    ENST_dict_tmp = {} # {ENSG: ENST_set, ...}
    vntHet_set = set() # variants to write in output -> {(i, vntHet_obj), ...}
                       # i is to track and keep variants order as they are read from input

    # Get idx for ENST and ENSG
    ENSG_idx = vcf_obj.header.get_tag_field_idx(VEPtag, 'Gene')
    ENST_idx = vcf_obj.header.get_tag_field_idx(VEPtag, 'Feature')

    # Get idx for SpliceAI, CLINVAR and VEP IMPACT
    if is_impct:
        try:
            IMPCT_idx = vcf_obj.header.get_tag_field_idx(VEPtag, 'IMPACT')
            is_IMPACT = True
        except Exception: # missing IMPACT, IMPCT_idx will point to Consequence
            try:
                IMPCT_idx = vcf_obj.header.get_tag_field_idx(VEPtag, 'Consequence')
            except Exception:
                sys.exit('\nERROR in VCF structure: either IMPACT or Consequence field in VEP is necessary to assign "--impact"\n')
            #end try
        #end try
        try:
            if SpliceAItag: # single tag has been specified
                tag, idx = vcf_obj.header.check_tag_definition(SpliceAItag)
                SpAItag_list.append(tag)
                SpAI_idx_list.append(idx)
            else: # search for delta scores as default
                for DStag in DStags:
                    tag, idx = vcf_obj.header.check_tag_definition(DStag)
                    SpAItag_list.append(tag)
                    SpAI_idx_list.append(idx)
                #end for
            #end if
            is_SpAI = True
        except Exception: is_SpAI = False
        #end try
        try:
            CLNSIGtag, CLNSIG_idx = vcf_obj.header.check_tag_definition('CLNSIG')
            is_CLNSIG = True
        except Exception: is_CLNSIG = False
        #end try
    #end if

    # Get trio IDs
    # if len(args['trio']) > 3:
    #     sys.exit('\nERROR in parsing arguments: too many sample IDs provided for trio\n')
    if len(args['trio']) == 3:
        is_family = True
    #end if
    ID_list = args['trio'][:3] # [proband_ID, parent_ID, parent_ID]

    # Reading variants
    analyzed = 0
    for c, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(c + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM):
        #     continue
        # #end if
        analyzed += 1

        # Reset data structures
        ENST_dict_tmp = {}
        IMPCT_dict_tmp = {}

        # Creating VariantHet object
        vntHet_obj = VariantHet(vnt_obj, c)
        if not filter_cmpHet: # if not filter, all variants are added to vntHet_set here
                              # if filter, no variant is added here to vntHet_set,
                              # compound heterozygous variants will be added after pairing
            vntHet_set.add((vntHet_obj.i, vntHet_obj))
        #end if

        # Check proband_ID genotype
        if vnt_obj.get_genotype_value(ID_list[0], 'GT').replace('|', '/') not in ['0/1', '1/0']:
            continue # go next if is not 0/1
        #end if
        # Check parent_ID genotypes
        if is_family:
            # go next if shared by both parents
            GT_1 = vnt_obj.get_genotype_value(ID_list[1], 'GT').replace('|', '/')
            GT_2 = vnt_obj.get_genotype_value(ID_list[2], 'GT').replace('|', '/')
            if GT_1 in ['0/1', '1/0'] and GT_2 in ['0/1', '1/0']:
                continue
            #end if
        #end if

        # Get transcripts and genes information from VEP
        ENSG_list = VEP_field(vnt_obj, ENSG_idx, VEPtag)
        ENST_list = VEP_field(vnt_obj, ENST_idx, VEPtag)

        # Assign transcripts to genes
        for ENSG, ENST in zip(ENSG_list, ENST_list):
            if ENSG and ENST:
                ENST_dict_tmp.setdefault(ENSG, set())
                ENST_dict_tmp[ENSG].add(ENST)
            #end if
        #end for

        # Assign variant to genes if VEP
        if ENST_dict_tmp:
            # Assign variant to genes and update transcripts for variant
            for ENSG, ENST_set in ENST_dict_tmp.items():
                ENSG_dict.setdefault(ENSG, [])
                ENSG_dict[ENSG].append(vntHet_obj)
                vntHet_obj.add_ENST(ENSG, ENST_set)
            #end for
        #end if

        # Add impact information if is_impct
        if is_impct:
            # VEP
            IMPCT_list = VEP_field(vnt_obj, IMPCT_idx, VEPtag)
            IMPCT_encoded = encode_IMPACT(IMPCT_list, VEP_encode, IMPCT_encode, is_IMPACT, sep)
            for i, (ENSG, IMPCT) in enumerate(zip(ENSG_list, IMPCT_encoded)):
                if ENSG and IMPCT:
                    IMPCT_dict_tmp.setdefault(ENSG, set())
                    IMPCT_dict_tmp[ENSG].add(IMPCT)
                    vntHet_obj.add_ENST_IMPCT(ENST_list[i], IMPCT)
                #end if
            #end for
            if IMPCT_dict_tmp:
                for ENSG, IMPCT_set in IMPCT_dict_tmp.items():
                    vntHet_obj.add_ENSG_IMPCT(ENSG, IMPCT_set)
                #end for
            #end if
            # SpliceAI
            if is_SpAI:
                # if SpliceAI is within VEP
                # fetching only the first transcript
                # expected the same scores for all transcripts
                SpAI_vals = []
                for i, SpAItag in enumerate(SpAItag_list):
                    SpAI_val = get_tag_idx(vnt_obj, SpAItag, SpAI_idx_list[i])
                    # if SpliceAI is with VEP and is at the end of Format
                    # need to remove , that separate next transcript
                    try: SpAI_vals.append(float(SpAI_val.split(',')[0]))
                    except Exception:
                        break
                    #end try
                #end for
                if SpAI_vals:
                    max_SpAI_vals = max(SpAI_vals)
                    if max_SpAI_vals >= 0.2:
                        vntHet_obj.add_SpAI(max_SpAI_vals)
                    #end if
                #end if
            #end if
            # CLINVAR
            if is_CLNSIG:
                # if CLNSIG is within VEP
                # fetching only the first transcript
                # expected the same CLNSIG for all transcripts
                CLNSIG_val = get_tag_idx(vnt_obj, CLNSIGtag, CLNSIG_idx)
                if CLNSIG_val:
                    vntHet_obj.add_CLINVAR(CLNSIG_val, CLNSIG_encode)
                #end if
            #end if
        #end if
    #end for

    # Pairing variants
    sys.stderr.write('\n')
    n = len(ENSG_dict)
    for n_i, (ENSG, vntHet_list) in enumerate(ENSG_dict.items()):
        if is_verbose:
            sys.stderr.write('\rPairing variants... {:.0f}%'.format(float(n_i)/n*100))
            sys.stderr.flush()
        #end if
        p, l = 0, len(vntHet_list)
        while p < l:
            vntHet_obj = vntHet_list[p]
            for i, vntHet_obj_i in enumerate(vntHet_list):
                if i != p:
                    # if parents information,
                    # check genotypes to confirm is compound het or not
                    if is_comHet(vntHet_obj, vntHet_obj_i, ID_list, allow_undef):
                        vntHet_obj.add_pair(vntHet_obj_i, ENSG, phase(vntHet_obj, vntHet_obj_i, ID_list), sep, is_impct, IMPCT_decode, test)
                        # Add vntHet to set to write since there is at least one pair
                        vntHet_set.add((vntHet_obj.i, vntHet_obj))
                    #end if
                #end if
            #end for
            p += 1
        #end while
    #end for
    if is_verbose:
        sys.stderr.write('\rPairing variants... {0}%'.format(100))
        sys.stderr.flush()
    #end if

    # Writing output
    sys.stderr.write('\n\n...Writing results for ' + str(analyzed) + ' analyzed variants out of ' + str(c + 1) + ' total variants\n')
    sys.stderr.flush()

    # Order and write variants to output file
    for _, vntHet_obj in sorted(vntHet_set, key=lambda x: x[0]):
        fo.write(vntHet_obj.to_string())
        update_stats(vntHet_obj, stat_dict, sep, is_impct)
    #end for

    # Print summary
    fs = open(args['outputfile'] + '.summary', 'w')
    fj = open(args['outputfile'] + '.json', 'w')

    # Get stats as json
    stat_json = to_json(stat_dict, is_impct)

    # Write to file
    print_stats(stat_json, fs, is_impct)
    json.dump(stat_json, fj, indent=2, sort_keys=True)

    # Close buffers
    fo.close()
    fs.close()
    fj.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    granite
#        Entry point for granite from the command line
#
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import argparse
# Tools
from granite import _version
from granite import novoCaller
from granite import comHet
from granite import blackList
from granite import whiteList
from granite import mpileupCounts
from granite import toBig
from granite import rckTar
from granite import cleanVCF
from granite import qcVCF
from granite import SVqcVCF
from granite import validateVCF
from granite import toPED
from granite import geneList


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    runner
#################################################################
def main():
    ''' command line wrapper around available tools '''
    # Adding parser and subparsers
    parser = argparse.ArgumentParser(prog='granite', description='granite ({0}) is a collection of software to work with genomic variants. The suite provides inheritance mode callers and utilities to filter and refine variants called by other methods in VCF format'.format(_version.__version__))
    subparsers = parser.add_subparsers(dest='func', metavar="<command>")

    # Add novoCaller to subparsers
    novoCaller_parser = subparsers.add_parser('novoCaller', description='Bayesian de novo variant caller',
                                                help='Bayesian de novo variant caller')

    novoCaller_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    novoCaller_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)
    novoCaller_parser.add_argument('-u', '--unrelatedfiles', help='TSV index file containing SampleID<TAB>Path/to/file for unrelated files used to train the model (BAM or bgzip and tabix indexed RCK)', type=str, required=True)
    novoCaller_parser.add_argument('-t', '--triofiles', help='TSV index file containing SampleID<TAB>Path/to/file for family files, the PROBAND must be listed as FIRST (BAM or bgzip and tabix indexed RCK)', type=str, required=True)
    novoCaller_parser.add_argument('--ppthr', help='threshold to filter by posterior probabilty for de novo calls (>=) [0]', type=float, required=False)
    novoCaller_parser.add_argument('--afthr', help='threshold to filter by population allele frequency (<=) [1]', type=float, required=False)
    novoCaller_parser.add_argument('--afthr_unrelated', help='threshold to filter by allele frequency calculated among unrelated (<=) [1]', type=float, required=False)
    novoCaller_parser.add_argument('--aftag', help='TAG (TAG=<float>) or TAG field to be used to filter by population allele frequency', type=str, required=False)
    novoCaller_parser.add_argument('--bam', help='by default the program expect bgzip and tabix indexed RCK files for "--triofiles" and "--unrelatedfiles", add this flag if files are in BAM format instead (SLOWER)', action='store_true', required=False)
    novoCaller_parser.add_argument('--MQthr', help='(only with "--bam") minimum mapping quality for an alignment to be used (>=) [0]', type=int, required=False)
    novoCaller_parser.add_argument('--BQthr', help='(only with "--bam") minimum base quality for a base to be considered (>=) [0]', type=int, required=False)
    novoCaller_parser.add_argument('--ADthr', help='threshold to filter by alternate allele depth in parents. This will ignore and set to "0" the posterior probability for variants with a number of alternate reads in parents higher than specified value', type=int, required=False)
    novoCaller_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add comHet to subparsers
    comHet_parser = subparsers.add_parser('comHet', description='compound heterozygous variant caller',
                                                help='compound heterozygous variant caller')

    comHet_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    comHet_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)
    comHet_parser.add_argument('--trio', help='list of sample IDs for trio, PROBAND is required and must be listed FIRST (e.g. --trio PROBAND_ID [PARENT_ID] [PARENT_ID])', nargs='+', required=True)
    comHet_parser.add_argument('--VEPtag', help='by default the program will search for "CSQ" TAG (CSQ=<values>), use this parameter to specify a different TAG to be used (e.g. VEP)', type=str, required=False)
    comHet_parser.add_argument('--sep', help='by default the program uses "&" as separator for subfields in annotating VCF (e.g. ENST00000643759&ENST00000643774), use this parameter to specify a different separator to be used', type=str, required=False)
    comHet_parser.add_argument('--filter_cmpHet', help='by default the program returns all variants in the input VCF file. This flag will produce a shorter output containing only variants that are potential compound heterozygous', action='store_true', required=False)
    comHet_parser.add_argument('--allow_undef', help='by default the program ignores variants with undefined genotype in parents. This flag extends the output to include these cases', action='store_true', required=False)
    comHet_parser.add_argument('--SpliceAItag', help='by default the program will search for SpliceAI delta scores (DS_AG, DS_AL, DS_DG, DS_DL) to calculate the max delta score for the variant. If a max value is already defined, use this parameter to specify the TAG | TAG field to be used', type=str, required=False)
    comHet_parser.add_argument('--impact', help='use VEP "IMPACT" or "Consequence" terms to assign an impact to potential compound heterozygous. If available, SpliceAI and ClinVar "CLNSIG" information is used together with VEP', action='store_true', required=False)
    comHet_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add mpileupCounts to subparsers
    mpileupCounts_parser = subparsers.add_parser('mpileupCounts', description='samtools wrapper to calculate reads statistics for pileup at each position',
                                                    help='samtools wrapper to calculate reads statistics for pileup at each position')

    mpileupCounts_parser.add_argument('-i', '--inputfile', help='input file in BAM format', type=str, required=True)
    mpileupCounts_parser.add_argument('-o', '--outputfile', help='output file to write results as RCK format (TSV), use .rck as extension', type=str, required=True)
    mpileupCounts_parser.add_argument('-r', '--reference', help='reference file in FASTA format', type=str, required=True)
    mpileupCounts_parser.add_argument('--region', help='region to be analyzed [e.g. chr1:1-10000000, 1:1-10000000, chr1, 1], chromosome name must match the reference', type=str, required=False)
    mpileupCounts_parser.add_argument('--MQthr', help='minimum mapping quality for an alignment to be used (>=) [0]', type=int, required=False)
    mpileupCounts_parser.add_argument('--BQthr', help='minimum base quality for a base to be considered (>=) [13]', type=int, required=False)

    # Add blackList to subparsers
    blackList_parser = subparsers.add_parser('blackList', description='utility to blacklist and filter out variants from input VCF file based on positions set in BIG format file and/or population allele frequency',
                                                help='utility to blacklist and filter out variants from input VCF file based on positions set in BIG format file and/or population allele frequency')

    blackList_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    blackList_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)
    blackList_parser.add_argument('-b', '--bigfile', help='BIG format file with positions set for blacklist', type=str, required=False)
    blackList_parser.add_argument('--aftag', help='TAG (TAG=<float>) or TAG field to be used to filter by population allele frequency', type=str, required=False)
    blackList_parser.add_argument('--afthr', help='threshold to filter by population allele frequency (<=) [1]', type=float, required=False)
    blackList_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add whiteList to subparsers
    whiteList_parser = subparsers.add_parser('whiteList', description='utility to whitelist and select a subset of variants from input VCF file based on specified annotations and positions',
                                                help='utility to whitelist and select a subset of variants from input VCF file based on specified annotations and positions')

    whiteList_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    whiteList_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)
    whiteList_parser.add_argument('--SpliceAI', help='threshold to whitelist variants by SpliceAI delta scores value (>=)', type=float, required=False)
    whiteList_parser.add_argument('--SpliceAItag', help='by default the program will search for SpliceAI delta scores (DS_AG, DS_AL, DS_DG, DS_DL) to calculate the max delta score for the variant. If a max value is already defined, use this parameter to specify the TAG | TAG field to be used', type=str, required=False)
    whiteList_parser.add_argument('--CLINVAR', help='flag to whitelist all variants with a ClinVar entry [ALLELEID]', action='store_true', required=False)
    whiteList_parser.add_argument('--CLINVARonly', help='ClinVar "CLNSIG" terms or keywords to be saved. Sets for whitelist only ClinVar variants with specified terms or keywords', nargs='+', required=False)
    whiteList_parser.add_argument('--CLINVARtag', help='by default the program will search for ClinVar "ALLELEID" TAG, use this parameter to specify a different TAG to be used', type=str, required=False)
    whiteList_parser.add_argument('--VEP', help='use VEP "Consequence" annotations to whitelist exonic and relevant variants (removed by default variants in intronic, intergenic, or regulatory regions)', action='store_true', required=False)
    whiteList_parser.add_argument('--VEPtag', help='by default the program will search for "CSQ" TAG (CSQ=<values>), use this parameter to specify a different TAG to be used (e.g. VEP)', type=str, required=False)
    whiteList_parser.add_argument('--VEPrescue', help='additional terms to overrule removed flags to rescue and whitelist variants', nargs='+', required=False)
    whiteList_parser.add_argument('--VEPremove', help='additional terms to be removed', nargs='+', required=False)
    whiteList_parser.add_argument('--VEPsep', help='by default the program expects "&" as separator for subfields in VEP (e.g. intron_variant&splice_region_variant), use this parameter to specify a different separator to be used', type=str, required=False)
    whiteList_parser.add_argument('--BEDfile', help='BED format file with positions to whitelist', type=str, required=False)
    whiteList_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add cleanVCF to subparsers
    cleanVCF_parser = subparsers.add_parser('cleanVCF', description='utility to clean INFO field of input VCF file',
                                                help='utility to clean INFO field of input VCF file')

    cleanVCF_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    cleanVCF_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)
    cleanVCF_parser.add_argument('-t', '--tag', help='TAG to be removed from INFO field. Specify multiple TAGs as: "-t TAG -t TAG -t ..."', action='append', required=False)
    cleanVCF_parser.add_argument('--VEP', help='clean VEP "Consequence" annotations (removed by default terms for intronic, intergenic, or regulatory regions from annotations)', action='store_true', required=False)
    cleanVCF_parser.add_argument('--VEPtag', help='by default the program will search for "CSQ" TAG (CSQ=<values>), use this parameter to specify a different TAG to be used (e.g. VEP)', type=str, required=False)
    cleanVCF_parser.add_argument('--VEPrescue', help='additional terms to overrule removed flags to rescue annotations', nargs='+', required=False)
    cleanVCF_parser.add_argument('--VEPremove', help='additional terms to be removed from annotations', nargs='+', required=False)
    cleanVCF_parser.add_argument('--VEPsep', help='by default the program expects "&" as separator for subfields in VEP (e.g. intron_variant&splice_region_variant), use this parameter to specify a different separator to be used', type=str, required=False)
    cleanVCF_parser.add_argument('--SpliceAI', help='threshold to save intronic annotations, from VEP "Consequence", for variants by SpliceAI delta scores value (>=)', type=float, required=False)
    cleanVCF_parser.add_argument('--SpliceAItag', help='by default the program will search for SpliceAI delta scores (DS_AG, DS_AL, DS_DG, DS_DL) to calculate the max delta score for the variant. If a max value is already defined, use this parameter to specify the TAG | TAG field to be used', type=str, required=False)
    cleanVCF_parser.add_argument('--filter_VEP', help='by default the program returns all variants in the input VCF file. This flag will drop the variants with no VEP annotations after the cleaning', action='store_true', required=False)
    cleanVCF_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add geneList to subparsers
    geneList_parser = subparsers.add_parser('geneList', description='utility to clean VEP annotations of input VCF file using a list of genes',
                                                help='utility to clean VEP annotations of input VCF file using a list of genes')

    geneList_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    geneList_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)
    geneList_parser.add_argument('-g', '--geneslist', help='text file listing ensembl gene (ENSG) IDs for all genes to save annotations for, IDs must be listed as a column', type=str, required=True)
    geneList_parser.add_argument('--VEPtag', help='by default the program will search for "CSQ" TAG (CSQ=<values>), use this parameter to specify a different TAG to be used (e.g. VEP)', type=str, required=False)
    geneList_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add toBig to subparsers
    toBig_parser = subparsers.add_parser('toBig', description='utility that converts counts from bgzip and tabix indexed RCK format into BIG format. Positions are "called" by reads counts or allelic balance for single or multiple files (joint calls) in specified regions',
                                                help='utility that converts counts from bgzip and tabix indexed RCK format into BIG format. Positions are "called" by reads counts or allelic balance for single or multiple files (joint calls) in specified regions')

    toBig_parser.add_argument('-f', '--file', help='file to be used to call positions. To do joint calling specify multiple files as: "-f file_1 -f file_2 -f ...". Expected bgzip and tabix indexed RCK file', action='append', required=True)
    toBig_parser.add_argument('-o', '--outputfile', help='output file to write results as BIG format (binary hdf5), use .big as extension', type=str, required=True)
    toBig_parser.add_argument('-r', '--regionfile', help='file containing regions to be used [e.g. chr1:1-10000000, 1:1-10000000, chr1, 1] listed as a column, chromosomes names must match the reference', type=str, required=True)
    toBig_parser.add_argument('-c', '--chromfile', help='chrom.sizes file containing chromosomes size information', type=str, required=True)
    toBig_parser.add_argument('--ncores', help='number of cores to be used if multiple regions are specified [1]', type=int, required=False)
    toBig_parser.add_argument('--fithr', help='minimum number of files with at least "--rdthr" for the alternate allele or having the variant, "calls" by allelic balance, to jointly "call" position (>=)', type=int, required=True)
    toBig_parser.add_argument('--rdthr', help='minimum number of alternate reads to count the file in "--fithr", if not specified "calls" are made by allelic balance (>=)', type=int, required=False)
    toBig_parser.add_argument('--abthr', help='minimum percentage of alternate reads compared to reference reads to count the file in "--fithr" when "calling" by allelic balance (>=) [15]', type=int, required=False)

    # Add rckTar to subparsers
    rckTar_parser = subparsers.add_parser('rckTar', description='utility to create a tar archive from bgzip and tabix indexed RCK files. Creates an index file for the archive',
                                                help='utility to create a tar archive from bgzip and tabix indexed RCK files. Creates an index file for the archive')

    rckTar_parser.add_argument('-t', '--ttar', help='target tar to write results, use .rck.tar as extension', type=str, required=True)
    rckTar_parser.add_argument('-f', '--file', help='file to be archived. Specify multiple files as: "-f SampleID_1.rck.gz -f SampleID_2.rck.gz -f ...". Files order is maintained while creating the index', action='append', required=True)

    # Add qcVCF to subparsers
    qcVCF_parser = subparsers.add_parser('qcVCF', description='utility to create a report of different metrics calculated for input VCF file',
                                                help='utility to create a report of different metrics calculated for input VCF file')

    qcVCF_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    qcVCF_parser.add_argument('-o', '--outputfile', help='output file to write results as JSON, use .json as extension', type=str, required=True)
    qcVCF_parser.add_argument('-p', '--pedigree', help='pedigree information, either as JSON file or JSON representation as string', type=str, required=True)
    qcVCF_parser.add_argument('--samples', help='list of sample IDs to get stats for (e.g. --samples SampleID_1 [SampleID_2] ...)', nargs='+', required=True)
    qcVCF_parser.add_argument('--ti_tv', help='add transition-transversion ratio and statistics on substitutions to report', action='store_true', required=False)
    qcVCF_parser.add_argument('--trio_errors', help='add statistics on mendelian errors based on trio to report', action='store_true', required=False)
    qcVCF_parser.add_argument('--het_hom', help='add heterozygosity ratio and statistics on zygosity to report', action='store_true', required=False)
    qcVCF_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add SVqcVCF to subparsers
    SVqcVCF_parser = subparsers.add_parser('SVqcVCF', description='utility to create a report of different metrics calculated for input SV VCF file',
                                                help='utility to create a report of different metrics calculated for input SV VCF file')

    SVqcVCF_parser.add_argument('-i', '--inputfile', help='input SV VCF file', type=str, required=True)
    SVqcVCF_parser.add_argument('-o', '--outputfile', help='output file to write results as JSON, use .json as extension', type=str, required=True)
    SVqcVCF_parser.add_argument('--samples', help='list of sample IDs to get stats for (e.g. --samples SampleID_1 [SampleID_2] ...)', nargs='+', required=True)
    SVqcVCF_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add validateVCF to subparsers
    validateVCF_parser = subparsers.add_parser('validateVCF', description='utility to calculate error models for input VCF file using pedigree information',
                                                    help='utility to calculate error models for input VCF file using pedigree information')

    validateVCF_parser.add_argument('-i', '--inputfile', help='input VCF file', type=str, required=True)
    validateVCF_parser.add_argument('-o', '--outputfile', help='output file to write results as JSON, use .json as extension', type=str, required=True)
    validateVCF_parser.add_argument('-p', '--pedigree', help='pedigree information, either as JSON file or JSON representation as string. It is possible to specify multiple pedigrees to load as list (e.g. --pedigree pedigree_1 [pedigree_2] ...)', nargs='+', required=True)
    validateVCF_parser.add_argument('--anchor', help='sample ID to be used as anchor in pedigree to build family. It is possible to specify multiple sample IDs as list. If multiple pedigrees are specified in "--pedigree", anchors are positionally matched to corresponding pedigree', nargs='+', required=True)
    validateVCF_parser.add_argument('--het', help='sample ID to be used to calculate error model for heterozygous mutations. It is possible to specify multiple sample IDs as list. Each sample ID must correspond to anchor specified in "--anchor"', nargs='+', required=False)
    validateVCF_parser.add_argument('--novo', help='sample ID to be used to calculate error model for de novo mutations. Must correspond to anchor specified in "--anchor". Requires posterior probability from novoCaller', type=str, required=False)
    validateVCF_parser.add_argument('--type', help='by default error models are calculated only for SNV. It is possible to specify different types of variant to use (SNV, INS, DEL, MNV, MAV) as list (e.g. --type SNV INS DEL)', nargs='+', required=False)
    validateVCF_parser.add_argument('--verbose', help='show progress status in terminal', action='store_true', required=False)

    # Add toPED to subparsers
    toPED_parser = subparsers.add_parser('toPED', description='utility to convert pedigree in JSON format to PED format',
                                                    help='utility to convert pedigree in JSON format to PED format')

    toPED_parser.add_argument('-p', '--pedigree', help='pedigree information, either as JSON file or JSON representation as string', type=str, required=True)
    toPED_parser.add_argument('-o', '--outputfile', help='output file to write results as PED, use .ped as extension', type=str, required=True)
    toPED_parser.add_argument('--family', help='family ID to be used [FAM]', type=str, required=False)

    # Add mergeVCF to subparsers
    # mergeVCF_parser = subparsers.add_parser('mergeVCF', description='utility to ',
    #                                             help='')
    # mergeVCF_parser.add_argument('-f', '--file', help='input VCF file to merge. Specify multiple files as: "-f file_1 -f file_2 -f ..."', action='append', required=True)
    # mergeVCF_parser.add_argument('-o', '--outputfile', help='output file to write results as VCF, use .vcf as extension', type=str, required=True)

    # Subparsers map
    subparser_map = {
                    'novoCaller': novoCaller_parser,
                    'comHet': comHet_parser,
                    'blackList': blackList_parser,
                    'whiteList': whiteList_parser,
                    'cleanVCF': cleanVCF_parser,
                    'geneList': geneList_parser,
                    'mpileupCounts': mpileupCounts_parser,
                    'toBig': toBig_parser,
                    'rckTar': rckTar_parser,
                    'qcVCF': qcVCF_parser,
                    'SVqcVCF': SVqcVCF_parser,
                    'validateVCF': validateVCF_parser,
                    'toPED': toPED_parser
                    }

    # Checking arguments
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    elif len(sys.argv) == 2:
        if sys.argv[1] in subparser_map:
            subparser_map[sys.argv[1]].print_help(sys.stderr)
            sys.exit(1)
        else:
            parser.print_help(sys.stderr)
            sys.exit(1)
        #end if
    #end if
    args = vars(parser.parse_args())

    # Call the right tool
    if args['func'] == 'novoCaller':
        novoCaller.main(args)
    elif args['func'] == 'comHet':
        comHet.main(args)
    elif args['func'] == 'blackList':
        blackList.main(args)
    elif args['func'] == 'whiteList':
        whiteList.main(args)
    elif args['func'] == 'mpileupCounts':
        mpileupCounts.main(args)
    elif args['func'] == 'toBig':
        toBig.main(args)
    elif args['func'] == 'rckTar':
        rckTar.main(args)
    elif args['func'] == 'cleanVCF':
        cleanVCF.main(args)
    elif args['func'] == 'geneList':
        geneList.main(args)
    elif args['func'] == 'qcVCF':
        qcVCF.main(args)
    elif args['func'] == 'SVqcVCF':
        SVqcVCF.main(args)
    elif args['func'] == 'validateVCF':
        validateVCF.main(args)
    elif args['func'] == 'toPED':
        toPED.main(args)
    #end if
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    geneList
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    ENSG_set, ENSG_idx = set(), 0
    VEPtag = args['VEPtag'] if args['VEPtag'] else 'CSQ'
    is_verbose = True if args['verbose'] else False

    # Read genes list
    with open(args['geneslist']) as fi:
        for line in fi:
            line = line.rstrip()
            if line: ENSG_set.add(line)
            #end if
        #end for
    #end with

    # Buffers
    fo = open(args['outputfile'], 'w', encoding='utf-8')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Writing header
    vcf_obj.write_header(fo)

    # Get ENSG (Gene) index in VEP
    ENSG_idx = vcf_obj.header.get_tag_field_idx(VEPtag, 'Gene')

    # Reading variants and writing passed
    analyzed = 0
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM):
        #     continue
        # #end if
        analyzed += 1

        # Apply genes list and clean VEP
        VEP_clean = clean_VEP_byfield(vnt_obj, ENSG_idx, ENSG_set, VEPtag)

        # Remove old VEP
        vnt_obj.remove_tag_info(VEPtag)

        # Add cleaned VEP if any
        if VEP_clean:
            vnt_obj.add_tag_info('{0}={1}'.format(VEPtag, VEP_clean))
        #end if

        # Write variant
        vcf_obj.write_variant(fo, vnt_obj)
    #end for
    sys.stderr.write('\n\n...Wrote results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    cleanVCF
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser
# shared_vars
from granite.lib.shared_vars import VEPremove
from granite.lib.shared_vars import VEPSpliceAI
from granite.lib.shared_vars import DStags


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    VEPrescue, consequence_idx = set(), 0
    # VEPremove = {...} -> import from shared_vars
    # VEPSpliceAI = {...} -> import from shared_vars
    # DStags = {...} -> import from shared_vars
    SpAItag_list, SpAI_idx_list = [], []
    is_VEP = True if args['VEP'] else False
    is_filter_VEP = True if args['filter_VEP'] else False
    VEPtag = args['VEPtag'] if args['VEPtag'] else 'CSQ'
    VEPsep = args['VEPsep'] if args['VEPsep'] else '&'
    SpliceAI_thr = float(args['SpliceAI']) if args['SpliceAI'] else 0.
    SpliceAItag = args['SpliceAItag'] # default None
    is_SpAI = False
    is_verbose = True if args['verbose'] else False

    # Buffers
    fo = open(args['outputfile'], 'w', encoding='utf-8')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Clean header definitions in INFO block for specified tags
    if args['tag']:
        for tag in args['tag']:
            vcf_obj.header.remove_tag_definition(tag, 'INFO')
        #end for
    #end if

    # Writing header
    vcf_obj.write_header(fo)

    # VEP
    if is_VEP:
        consequence_idx = vcf_obj.header.get_tag_field_idx(VEPtag, 'Consequence')
        if args['VEPrescue']: VEPrescue = {term for term in args['VEPrescue']}
        #end if
        if args['VEPremove']: VEPremove.update({term for term in args['VEPremove']})
        #end if
    elif args['VEPrescue'] or args['VEPremove']:
        sys.exit('\nERROR in parsing arguments: specify the flag "--VEP" to filter by VEP annotations to apply rescue terms or remove additional terms\n')
    #end if

    # SpliceAI
    if SpliceAI_thr:
        if SpliceAItag: # single tag has been specified
            tag, idx = vcf_obj.header.check_tag_definition(SpliceAItag)
            SpAItag_list.append(tag)
            SpAI_idx_list.append(idx)
        else: # search for delta scores as default
            for DStag in DStags:
                tag, idx = vcf_obj.header.check_tag_definition(DStag)
                SpAItag_list.append(tag)
                SpAI_idx_list.append(idx)
            #end for
        #end if
    #end if

    # Reading variants and writing passed
    analyzed = 0
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM):
        #     continue
        # #end if
        analyzed += 1

        # Clean tags if specified
        if args['tag']:
            for tag in args['tag']:
                vnt_obj.remove_tag_info(tag)
            #end for
        #end if

        # is_SpAI reset
        is_SpAI = False

        # Check SpliceAI
        if SpliceAI_thr:
            if check_spliceAI(vnt_obj, SpAI_idx_list, SpAItag_list, SpliceAI_thr):
                is_SpAI = True
            #end if
        #end if

        # Clean VEP
        if is_VEP:
            # Get cleaned VEP
            if is_SpAI:
                VEP_clean = clean_VEP(vnt_obj, consequence_idx, VEPremove, VEPrescue.union(VEPSpliceAI), VEPtag, VEPsep)
            else:
                VEP_clean = clean_VEP(vnt_obj, consequence_idx, VEPremove, VEPrescue, VEPtag, VEPsep)
            #end if
            # Remove old VEP
            vnt_obj.remove_tag_info(VEPtag)
            # Add cleaned VEP if any
            if VEP_clean:
                vnt_obj.add_tag_info('{0}={1}'.format(VEPtag, VEP_clean))
            else: # check if is filter_VEP
                if is_filter_VEP:
                    continue
                #end if
            #end if
        #end if

        # Write variant
        vcf_obj.write_variant(fo, vnt_obj)
    #end for
    sys.stderr.write('\n\n...Wrote results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    toBig
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import h5py
import multiprocessing
import bitarray
from multiprocessing import Pool
# shared_functions as *
from granite.lib.shared_functions import *


#################################################################
#
#    FUNCTIONS
#
#################################################################
def check_pos(rdthr, abthr, cov, ref_fw, ref_rv, alt_fw, alt_rv, ins_fw, ins_rv, del_fw, del_rv):
    ''' check if position can be called as snv, insertion or delition.
    absolute reads counts or allelic balance can be used alternatively '''
    is_snv, is_ins, is_del = False, False, False
    if not rdthr:
        is_snv = __routine_allbal(abthr, cov, alt_fw, alt_rv)
        is_ins = __routine_allbal(abthr, cov, ins_fw, ins_rv)
        is_del = __routine_allbal(abthr, cov, del_fw, del_rv)
    else:
        is_snv = __routine_reads(rdthr, alt_fw, alt_rv)
        is_ins = __routine_reads(rdthr, ins_fw, ins_rv)
        is_del = __routine_reads(rdthr, del_fw, del_rv)
    #end if
    return is_snv, is_ins, is_del
#end def

def __routine_allbal(abthr, cov, alt_fw, alt_rv):
    ''' check if position can be called as alternate based on allelic balance '''
    alt_tot = alt_fw + alt_rv
    prc_alt = alt_tot / cov * 100
    if prc_alt >= abthr:
        return True
    #end if
    return False
#end def

def __routine_reads(rdthr, alt_fw, alt_rv):
    ''' check if position can be called as alternate based on absolute
    read counts '''
    if alt_fw + alt_rv >= rdthr:
        return True
    #end if
    return False
#end def

def bitarrays_toHDF5(filename):
    ''' write bitarrays to file in HDF5 format '''
    fo = h5py.File(filename, 'w')
    # bitarrays is in global scope
    for chr in bitarrays:
        # Packing and writing snv
        fo[chr + '_snv'] = bitarrays[chr]['snv']
        # Packing and writing ins
        fo[chr + '_ins'] = bitarrays[chr]['ins']
        # Packing and writing del
        fo[chr + '_del'] = bitarrays[chr]['del']
    #end for
    fo.close()
#end def

def run_region(files, fithr, rdthr, abthr, region):
    ''' '''
    snv, ins, dele = [], [], []
    # Opening buffers to region
    buffers = [tabix_IT(filename, region) for filename in files]
    # Reading from buffers and processing
    bams_snv, bams_ins, bams_del = 0, 0, 0
    tmp_chr, tmp_pos = '', 0
    while True:
        bams_snv, bams_ins, bams_del = 0, 0, 0 # new position
                                               # reset bams counts
        # Check first bam
        try:
            line_split = next(buffers[0])
            chr = line_split[0]
            pos, cov, ref_fw, ref_rv, alt_fw, alt_rv, \
                ins_fw, ins_rv, del_fw, del_rv = map(int, line_split[1:])
        except Exception: break
        #end try
        tmp_chr, tmp_pos = chr, pos
        # Check position and update bams counts
        is_snv, is_ins, is_del = \
            check_pos(rdthr, abthr, cov, ref_fw, ref_rv, alt_fw, alt_rv, ins_fw, ins_rv, del_fw, del_rv)
        bams_snv += int(is_snv); bams_ins += int(is_ins); bams_del += int(is_del)
        # Check ramaining bams
        for i, buffer in enumerate(buffers[1:]):
            line_split = next(buffer)
            chr = line_split[0]
            pos, cov, ref_fw, ref_rv, alt_fw, alt_rv, \
                ins_fw, ins_rv, del_fw, del_rv = map(int, line_split[1:])
            # Check consistency among the files
            if tmp_chr != chr or tmp_pos != pos:
                raise IndexError('\nERROR in file: position {0}:{1} in file {2} is not consistent with other files\n'
                        .format(chr, pos, files[i+1]))
            #end if
            # Check position and update bams counts
            is_snv, is_ins, is_del = \
                check_pos(rdthr, abthr, cov, ref_fw, ref_rv, alt_fw, alt_rv, ins_fw, ins_rv, del_fw, del_rv)
            bams_snv += int(is_snv); bams_ins += int(is_ins); bams_del += int(is_del)
        #end for
        # Check thresholds
        if bams_snv >= fithr:
            snv.append((tmp_chr, tmp_pos))
        #end if
        if bams_ins >= fithr:
            ins.append((tmp_chr, tmp_pos))
        #end if
        if bams_del >= fithr:
            dele.append((tmp_chr, tmp_pos))
        #end if
    #end while
    return [snv, ins, dele]
#end def

def update_bitarrays(snv_ins_dele):
    ''' '''
    # bitarrays is in global scope
    for tmp_chr, tmp_pos in snv_ins_dele[0]:
        bitarrays[tmp_chr]['snv'][tmp_pos] = True
    #end for
    for tmp_chr, tmp_pos in snv_ins_dele[1]:
        bitarrays[tmp_chr]['ins'][tmp_pos] = True
    #end for
    for tmp_chr, tmp_pos in snv_ins_dele[2]:
        bitarrays[tmp_chr]['del'][tmp_pos] = True
    #end for

#################################################################
#    runner
#################################################################
def main(args):
    ''' '''

    global bitarrays

    # Variables
    fithr = int(args['fithr'])
    rdthr = int(args['rdthr']) if args['rdthr'] else 0
    abthr = int(args['abthr']) if args['abthr'] else 15
    ncores = int(args['ncores']) if args['ncores'] else 1
    files = args['file']

    # Data structures
    chr_length = {}
    regions = []
    bitarrays = {}

    # Reading chrom.sizes file
    with open(args['chromfile']) as fi:
        for line in fi:
            line = line.rstrip()
            if line:
                chr, length = line.split('\t')
                chr_length.setdefault(chr, int(length))
            #end if
        #end for
    #end with

    # Getting regions
    with open(args['regionfile']) as fi:
        for line in fi:
            line = line.rstrip() # line is a region
            if line:
                check_region(line, chr_length)
                regions.append(line)
            #end if
        #end for
    #end with

    # Initializing bitarrays data structure
    for chr, length in chr_length.items(): # +1 to index positions in bitarrays by 1
        bitarrays.setdefault(chr, {'snv': bitarray.bitarray(length + 1),
                                   'ins': bitarray.bitarray(length + 1),
                                   'del': bitarray.bitarray(length + 1)})
        bitarrays[chr]['snv'].setall(False)
        bitarrays[chr]['ins'].setall(False)
        bitarrays[chr]['del'].setall(False)
    #end for

    # Multiprocessing
    pool = Pool(ncores)
    for region in regions:
        p = pool.apply_async(run_region, (files, fithr, rdthr, abthr, region, ), callback=update_bitarrays)
        p.get()
    pool.close()
    pool.join()

    # Writing bitarrays to files
    bitarrays_toHDF5(args['outputfile'])
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if
